package pv260.customeranalysis;

import static com.googlecode.catchexception.CatchException.catchException;
import static java.util.Arrays.asList;
import static org.assertj.core.api.Assertions.assertThat;

import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InOrder;
import pv260.customeranalysis.entities.Customer;
import pv260.customeranalysis.entities.Offer;
import pv260.customeranalysis.entities.Product;
import pv260.customeranalysis.exceptions.CantUnderstandException;
import pv260.customeranalysis.exceptions.GeneralException;
import pv260.customeranalysis.interfaces.AnalyticalEngine;
import pv260.customeranalysis.interfaces.ErrorHandler;
import pv260.customeranalysis.interfaces.NewsList;
import pv260.customeranalysis.interfaces.Storage;

import static org.mockito.Matchers.isA;

import static org.mockito.Matchers.any;
import static org.mockito.Mockito.*;

public class CustomerAnalysisTest {
    ErrorHandler errorHandler;
    AnalyticalEngine engine1;
    Product product;
    Storage storage;
    NewsList newsList;

    @Before
    public void setUp() {
        errorHandler = mock(ErrorHandler.class);
        engine1 = mock(AnalyticalEngine.class);
        product = mock(Product.class);
        storage = mock(Storage.class);
        newsList = mock(NewsList.class);
    }


    /**
     * Verify the ErrorHandler is invoked when one of the AnalyticalEngine methods
     * throws exception and the exception is not re-thrown from the CustomerAnalysis.
     * The exception is passed to the ErrorHandler directly, not wrapped.
     */
    @Test
    public void testErrorHandlerInvokedWhenEngineThrows() throws GeneralException {
        when(engine1.interesetingCustomers(product)).thenThrow(new CantUnderstandException());

        CustomerAnalysis analysis = new CustomerAnalysis(asList(engine1), storage, newsList, errorHandler);
        catchException(() -> analysis.findInterestingCustomers(product));

        verify(errorHandler).handle(isA(CantUnderstandException.class));
    }

    /**
     * Verify that if first AnalyticalEngine fails by throwing an exception,
     * subsequent engines are tried with the same input.
     * Ordering of engines is given by their order in the List passed to
     * constructor of AnalyticalEngine
     */
    @Test
    public void testSubsequentEnginesTriedIfOneFails() throws GeneralException {
        AnalyticalEngine engine2 = mock(AnalyticalEngine.class);
        when(engine1.interesetingCustomers(product)).thenThrow(new CantUnderstandException());

        CustomerAnalysis analysis = new CustomerAnalysis(asList(engine1, engine2), storage, newsList, errorHandler);

        catchException(() -> analysis.findInterestingCustomers(product));
        verify(engine1).interesetingCustomers(product);
        verify(engine2).interesetingCustomers(product);
    }

    /**
     * Verify that as soon as the first AnalyticalEngine succeeds,
     * this result is returned as result and no subsequent
     * AnalyticalEngine is invoked for this input
     */
    @Test
    public void testNoMoreEnginesTriedAfterOneSucceeds() throws GeneralException {
        AnalyticalEngine engine2 = mock(AnalyticalEngine.class);

        CustomerAnalysis analysis = new CustomerAnalysis(asList(engine1, engine2), storage, newsList, errorHandler);

        catchException(() -> analysis.findInterestingCustomers(product));

        verify(engine2, never()).interesetingCustomers(product);
    }

    /**
     * Verify that once Offer is created for the Customer,
     * this order is persisted in the Storage before being
     * added to the NewsList
     * HINT: you might use mockito InOrder
     */
    @Test
    public void testOfferIsPersistedBefreAddedToNewsList() throws GeneralException {
        when(storage.find(Product.class, 1L)).thenReturn(product);
        when(product.getId()).thenReturn(1L);
        CustomerAnalysis analysis = spy(new CustomerAnalysis(asList(engine1), storage, newsList, errorHandler));
        //when(analysis.findInterestingCustomers(product)).thenReturn(asList(any(Customer.class)));
        doReturn(asList(mock(Customer.class))).when(analysis).findInterestingCustomers(product);

        catchException(() -> analysis.prepareOfferForProduct(1L));
        InOrder inOrder = inOrder(storage, newsList);

        inOrder.verify(storage).persist(any(Offer.class));
        inOrder.verify(newsList).sendPeriodically(any(Offer.class));
    }

    /**
     * Verify that Offer is created for every selected Customer for the given Product
     * test with at least two Customers selected by the AnalyticalEngine
     * HINT: you might use mockito ArgumentCaptor 
     */
    @Test
    public void testOfferContainsProductAndCustomer() throws GeneralException {
        when(storage.find(Product.class, 1L)).thenReturn(product);
        CustomerAnalysis analysis = spy(new CustomerAnalysis(asList(engine1), storage, newsList, errorHandler));
        Customer c1 = mock(Customer.class);
        Customer c2 = mock(Customer.class);

        doReturn(asList(c1, c2)).when(analysis).findInterestingCustomers(product);

        catchException(() -> analysis.prepareOfferForProduct(1L));

        ArgumentCaptor<Offer> captor = ArgumentCaptor.forClass(Offer.class);
        verify(storage, times(2)).persist(captor.capture());
    }

}
